#include "korg.hpp"
#include "korgtext.hpp"
#include <string.h>

char* bsqname;
FILE* bsqf;
BSQHEAD bsqhead;
KORGNAME name;

BSQDATA bsq;
BSQDATAADR bsqadr;

void dumpkbdattr(int i, KBDATTR& attr)
{
  printf("KBD%d   Program: %s  Octave: %d\n",
    i+1, place2str(attr.prog), attr.octave+1);
}

void dumpcontrol(int i, TRACK_CTRLDATA& trk)
{
  printf("TRACK %d CONTROL DATA\n", i+1);
  printf("Program:%s  Volume:%d Pan:%s CD SendLevel:%s Track Status:%s\n",
    place2str(trk.prog), trk.vol, pan2str(trk.pan),
    efflev2str(trk.cd_sendlevel, 0),
    trackstat2str(trk.trackstat));
  printf("Bendrange:%d Transpose:%d Detune:%d Protect:%s\n",
    trk.bendrange, trk.transpose, trk.detune, onoff2str(trk.protect));
  printf("Channel:%d Velocity:%d-%d Keys:%s-%s\n",
    trk.channel,
    trk.vel_wintop, trk.vel_winbottom,
    note2str(trk.key_wintop,-1), note2str(trk.key_winbottom, -1));
}

void dumpbsqdata(BSQDATA& bsq)
{
  printf("Arrangement %s  ", place2str(bsq.arrno));
  printf("Style %s Variation %d\n",
    place2str(bsq.styno),
    bsq.variation+1);
  printf("Tempo %d  Keyboardassign %s\n",
    bsq.tempo + 40-10,
    kbdassign2str(bsq.keybassign));
  printf("Chord scan low: %s  Chord scan high: %s Bass Inversion: %s\n",
    onoff2str(bsq.chord_scan_low),
    onoff2str(bsq.chord_scan_high),
    onoff2str(bsq.bass_invers));
  printf("Chord hold: %s  Chord latch: %s\n",
    onoff2str(bsq.chord_hold),
    onoff2str(bsq.chord_latch));
  dumpkbdattr(0, bsq.kbdattr[0]);
  dumpkbdattr(1, bsq.kbdattr[1]);
  printf("TRACK STATUS  Keyboard:%s  Control:%s Chord:%s\n",
     onoff2str(bsq.kbd_trkstatus),
     onoff2str(bsq.cntrl_trkstatus),
     onoff2str(bsq.chord_trkstatus));
  printf("AutoTempo:%d  Beat:%s Splitpoint:%s\n",
    bsq.autotempo+40-10,
    tact2str(bsq.beat),
    note2str(bsq.splitpoint, 2)
    );
  printf("Transpose:%d DynamicVelocity:%s\n",
    bsq.transpose,
    onoff2str(bsq.sw_dynvel));

  printf("Next BSQ: %d  AutoStart:%s\n",
    bsq.next_bsq,
    onoff2str(bsq.autostart));

  for (int t = 0; t < 8; t++)
    dumpcontrol(t, bsq.track[t]);
  printf("EFFECT PARAMETERS (not visible)\n");
/*
  EFFECTPARAM effects;
*/
}

int dumpnoteevent(NOTEEVENT& note)
{
    printf("  %5s %-3s  V%03d",
      loc2str(note.loc),
      note2str(note.key, -1),
      note.vel * 2);
    printf("  %s\n", len2str(note.len));
    return 1;
}

int dumpctrlevent(CTRLEVENT& ctrl)
{
  switch(ctrl.cmd)
  {
  case 0x08:
    {
    int v = (((PITCHEVENT*)&ctrl)->v_hi << 7) + ((PITCHEVENT*)&ctrl)->v_lo;

    if (v < 0x2000)
      v |= 0xe000U;
    else
      v &= 0x1fffU;
    printf("  %5s BEND %d\n",
      loc2str(((PITCHEVENT*)&ctrl)->loc),
      v);
    }
    break;
  case 0x10:
    printf("  %5s AFTT %03d\n",
      loc2str(((AFTTEVENT*)&ctrl)->loc),
      ((AFTTEVENT*)&ctrl)->v);
    break;
  case 0x18:
    printf("  %5s PROGRAM %s\n",
      loc2str(((PROGEVENT*)&ctrl)->loc),
      place2str(((PROGEVENT*)&ctrl)->prog));
    break;
  case 0x20:
    if (ctrl.control == 107)
      printf("TEMPO %03d\n", ctrl.value + 30);
    else
      printf("  %5s CTRL %03d %03d\n",
	loc2str(ctrl.loc),
	ctrl.control, ctrl.value);
    break;
  case 0x30:
    printf("BAR %d %s\n", ((BEATEVENT*)&ctrl)->index+1, tact2str(((BEATEVENT*)&ctrl)->tact));
    break;
  case 0x38:
    printf("END TRACK %d\n", ((ENDEVENT*)&ctrl)->bar);
    return -1;
  default:
    return 0;
  }
  return 1;
}

int dumpbseqevent(CONTROLEVENT& control)
{
  printf("  %5s ", loc2str(control.time));
  switch(control.control)
  {
  case 0:
    printf("ARRANGEMENT: %s\n", arr2str(control.v1));
    break;
  case 1:
    printf("STYLE: %s\n", style2str(control.v1));
    break;
  case 2:
    printf("STY.element: %s\n", styelement2str(control.v1));
    break;
  case 3:
    printf("KBD assign: %s\n", kbdassign2str(control.v1));
    break;
  case 4:
    printf("Chord Scan: %s\n", chordscan2str(control.v1));
    break;
  case 5:
    printf("Chord Hold: %s\n", onoff2str(control.v1));
    break;
  case 6:
    printf("Bass Inversion: %s\n", onoff2str(control.v1));
    break;
  case 7:
    printf("Transpose: %d\n", (signed char)control.v1);
    break;
  case 8:
    printf("Mute drum: %s\n", onoff2str(control.v1));
    break;
  case 9:
    printf("Mute perc: %s\n", onoff2str(control.v1));
    break;
  case 10:
    printf("Mute bass: %s\n", onoff2str(control.v1));
    break;
  case 11:
    printf("Mute acc1: %s\n", onoff2str(control.v1));
    break;
  case 12:
    printf("Mute acc2: %s\n", onoff2str(control.v1));
    break;
  case 13:
    printf("Mute acc3: %s\n", onoff2str(control.v1));
    break;
  case 14:
    printf("Drum.map: %d\n", control.v1);
    break;
  case 15:
    printf("KBD1.prog: %s\n", place2str(control.v1, control.v2));
    break;
  case 16:
    printf("KBD2.prog: %s\n", place2str(control.v1, control.v2));
    break;
  case 17:
    printf("KBD1.octave: %d\n", (signed char)control.v1);
    break;
  case 18:
    printf("KBD2.octave: %d\n", (signed char)control.v1);
    break;
  default:
    return 0;
  }
  return 1;
}

int dumpchordevent(CHORDEVENT& chord)
{
  printf("  %5s %s\n", loc2str(chord.time),
    chord2str(chord.chord, chord.tensionnote, chord.bass, chord.root)
  );
  return 1;
}

int dumpevent(BSQEVENT& ev, int trk)
{
  switch(ev.general[1] >> 6)
  {
  case 0:
  case 1:
    return dumpctrlevent(ev.ctrl);
  case 2:
    if (trk == 1)
      return dumpbseqevent(ev.control);
    else
      return dumpnoteevent(ev.note);
  case 3:
    if (trk == 2)
      return dumpchordevent(ev.chord);
    else
      return dumpnoteevent(ev.note);
  }
  return 0;
}

void dumpsequence(int trk)
{
BSQEVENT ev;
int i = 0, rt;
long pos;

  while (KORGREAD(bsqf, ev))
  {
    pos = ftell(bsqf) - 4;
    printf("%04lX: ", pos);
    if ((rt = dumpevent(ev, trk)) == 0)
    {
      printf("unknown event: %02X %02X %02X %02X\n",
	ev.general[0], ev.general[1], ev.general[2], ev.general[3]);
    }
    if (rt == -1)
      break;
    i++;
  }
}

int isemptytrack(int bseq, int t)
{
  if (t < 16)
    return bsqadr.adr[bseq][t] >= bsqadr.adr[bseq][t+1];
  if (bseq < 9)
    return bsqadr.adr[bseq][t] >= bsqadr.adr[bseq+1][0];
  return 1;
}

int isemptybseq(int bseq)
{
  if (bseq < 0 || bseq >= 10)
    return 1;
  for (int t = 0; t < 17; t++)
    if (!isemptytrack(bseq, t))
      return 0;
  return 1;
}

void dumpbsq()
{
int i, t;
long bsqofs = 0;

  bsqofs += sizeof(bsqhead);
  if (!bsqheader(bsqf, bsqhead))
  {
    fprintf(stderr, "%s is not a KORG bsq file\n", bsqname);
    return;
  }
  printf("\nKORG BACKINGSEQUENCE file %s\n\n", bsqname);

  if (sizeof(bsq) != 0xc3)
    putchar(7);
  for (i = 0; i < 10; i++)
  {
    bsqofs += sizeof(bsq);
    if (!KORGREAD(bsqf, bsq))
      return;
    printf("\nBSEQ%d %10.10s\n", i, bsq.name);
    dumpbsqdata(bsq);
  }
  bsqofs += sizeof(bsqadr);
  if (!KORGREAD(bsqf, bsqadr))
    return;
  printf("      KBD  CTRL CHORD                         ET1  ET2  ET3  ET4  ET5  ET6  ET7  ET8  TEMPO\n");
  for (i = 0; i < 10; i++)
  {
    printf("BSEQ%d ", i);
    for (t = 0; t < 17; t++)
      if (isemptytrack(i, t))
	printf("  -  ");
      else
	printf("%04lX ", (long(bsqadr.adr[i][t]  - bsqadr.adr[0][0]) << 4) + bsqofs);
    printf("\n");
  }
  printf("\n");
  for (i = 0; i < 10; i++)
  {
    if (isemptybseq(i))
    {
      printf("BSEQ %d not used\n", i);
      continue;
    }
    for (t = 0; t < 17; t++)
    {
       if (isemptytrack(i, t))
	 continue;
       printf("BSEQ %d  Track %s\n", i, bseqtrack(t));
       fseek(bsqf, (long(bsqadr.adr[i][t]  - bsqadr.adr[0][0]) << 4) + bsqofs, SEEK_SET);
       dumpsequence(t);
       printf("\n");
    }
  }
}

void main(int argc, char**argv)
{
  argc--; argv++;

  if (argc == 0)
    return;

  bsqname = *argv++; argc--;

  bsqf = fopen(bsqname, "rb");
  if (!bsqf)
  {
    perror(bsqname);
    return;
  }

  dumpbsq();

  fclose(bsqf);
}
