// (c) 1999 by G. Nagler
#ifndef __KORGCOMPATIBILITY_HPP
#define __KORGCOMPATIBILITY_HPP

#include <stdio.h>

class MidiParam
{
public:
  MidiParam();

  int hbank, lbankmin, lbankmax, progmin, progmax;
};

class MIDIParamDef
{
public:
  MIDIParamDef();

  int bank;
  int progmin, progmax;

  MidiParam midi;
};

class BankParam
{
public:
  BankParam();

  char name[10];
  int bank;
  int progmin, progmax;
  char title[10];
};

class CompatParam
{
public:
  CompatParam();

  int bank, prog;
  int bankX, progX;
};

class CompatOscParam
{
public:
  CompatOscParam();

  int osc, oscX;
};

class ProgramParam
{
public:
  ProgramParam();

  int bank, prog;
  char name[20];
};

class KorgIProgramDef
{
public:
  KorgIProgramDef(const char* filename);
  ~KorgIProgramDef();

  FILE* getf();

  const char* getName();

  int getBankParam(int bank, int prog, BankParam& bankinfo);
  int getBankParam(const char *name, BankParam& bankinfo);
  int getDrumBank(BankParam& bankinfo);
  int getDrumBank();
  int isUserProg(int bank, int prog);
  int getMIDIParam(int bank, int prog, MidiParam& param);
  int getMIDIInit(unsigned char* data, int maxlen);
  int getKeyboardParam(MidiParam& param, int &bank, int &prog, int needdrum = 0);

  int getCompatiblei3(int bank, int prog, int& banki3, int& progi3);
  int getCompatible(const char* keyboard, int bank, int prog, int& bankX, int& progX);
  int getReverseCompatible(const char* keyboard, int bankX, int progX, int& bank, int& prog);

  const char* getProgramName(int bank, int prog);
  int findProgram(const char* name, int& bank, int& prog);

  int getCompatiblei3Oscillator(int osc, int& osci3);
  int getCompatibleOscillator(const char* keyboard, int osc, int& oscX);
  int getReverseCompatibleOscillator(const char* keyboard, int oscX, int& osc);

private:
  FILE* f_;
  char keyboard_[10];

  int findsection(const char* s);
  void skipws(char* &s);
  int getnumberrange(char* &s, int& min, int& max);
  int getnumber(char*& s);
  int matchchar(char*& s, int c);

  int getmidiparamdef(char* line, MIDIParamDef& param);
  int getbankdef(char* &s, BankParam& bankinfo);
  int getcompatdef(char* &s, CompatParam& compat);
  int getprogramdef(char*& s, ProgramParam& proginfo);
  int getcompatoscdef(char* &s, CompatOscParam& compat);

};

int getCompatible(KorgIProgramDef& def, int bank, int prog, KorgIProgramDef& defx, int &bankx, int &progx);
int getCompatibleOscillator(KorgIProgramDef& def, int osc, KorgIProgramDef& defx, int& oscx);

#endif

